"""
Module defining a specific exception for the errors occuring while analyzing
an XML Schema.

.. autoclass:: XsdError
   :members:
   :undoc-members:
   :show-inheritance: 
"""

class XsdError(Exception):
    """
    Exception raised when finding an error while analyzing an XML Schema.
    """
    pass

